import sys
import ssl
import asyncio
import traceback
import threading
import random
import json

from contextlib import suppress
from urllib.parse import urlencode
from tqdm import tqdm
import certifi
import backoff
from opencage.geocoder import (
    OpenCageGeocode,
    OpenCageGeocodeError,
    _query_for_reverse_geocoding,
    floatify_latlng
)


class OpenCageBatchGeocoder():

    """ Called from command_line.py
        init() receives the parsed command line parameters
        geocode() receive an input and output CSV reader/writer and loops over the data
    """

    def __init__(self, options):
        self.options = options
        self.sslcontext = ssl.create_default_context(cafile=certifi.where())
        self.user_agent_comment = 'OpenCage CLI'
        self.write_counter = 1

    def __call__(self, *args, **kwargs):
        asyncio.run(self.geocode(*args, **kwargs))

    async def geocode(self, csv_input, csv_output):
        if not self.options.dry_run:
            test = await self.test_request()
            if test['error']:
                self.log(test['error'])
                return
            if test['free'] is True and self.options.workers > 1:
                sys.stderr.write("Free trial account detected. Resetting number of workers to 1.\n")
                self.options.workers = 1

        if self.options.headers:
            header_columns = next(csv_input, None)
            if header_columns is None:
                return

        queue = asyncio.Queue(maxsize=self.options.limit)

        read_warnings = await self.read_input(csv_input, queue)

        if self.options.dry_run:
            if not read_warnings:
                print('All good.')
            return

        if self.options.headers:
            csv_output.writerow(header_columns + self.options.add_columns)

        progress_bar = not (self.options.no_progress or self.options.quiet) and \
            tqdm(total=queue.qsize(), position=0, desc="Addresses geocoded", dynamic_ncols=True)

        tasks = []
        for _ in range(self.options.workers):
            task = asyncio.create_task(self.worker(csv_output, queue, progress_bar))
            tasks.append(task)

        # This starts the workers and waits until all are finished
        await queue.join()

        # All tasks done
        for task in tasks:
            task.cancel()

        if progress_bar:
            progress_bar.close()

    async def test_request(self):
        try:
            async with OpenCageGeocode(
                self.options.api_key,
                domain=self.options.api_domain,
                sslcontext=self.sslcontext,
                user_agent_comment=self.user_agent_comment
            ) as geocoder:
                result = await geocoder.geocode_async('Kendall Sq, Cambridge, MA', raw_response=True)

                free = False
                with suppress(KeyError):
                    free = result['rate']['limit'] == 2500

                return {'error': None, 'free': free}
        except Exception as exc:
            return {'error': exc}

    async def read_input(self, csv_input, queue):
        any_warnings = False
        for index, row in enumerate(csv_input):
            line_number = index + 1

            if len(row) == 0:
                self.log(f"Line {line_number} - Empty line")
                any_warnings = True
                row = ['']

            item = await self.read_one_line(row, line_number)
            if item['warnings'] is True:
                any_warnings = True
            await queue.put(item)

            if queue.full():
                break

        return any_warnings

    async def read_one_line(self, row, row_id):
        warnings = False

        if self.options.input_columns:
            input_columns = self.options.input_columns
        elif self.options.command == 'reverse':
            input_columns = [1, 2]
        else:
            input_columns = None

        if input_columns:
            address = []
            try:
                for column in input_columns:
                    # input_columns option uses 1-based indexing
                    address.append(row[column - 1])
            except IndexError:
                self.log(f"Line {row_id} - Missing input column {column} in {row}")
                warnings = True
        else:
            address = row

        if self.options.command == 'reverse':

            if len(address) != 2:
                self.log(
                    f"Line {row_id} - Expected two comma-separated values for reverse geocoding, got {address}")
            else:
                # _query_for_reverse_geocoding attempts to convert into numbers. We rather have it fail
                # now than during the actual geocoding
                try:
                    _query_for_reverse_geocoding(address[0], address[1])
                except BaseException:
                    self.log(
                        f"Line {row_id} - Does not look like latitude and longitude: '{address[0]}' and '{address[1]}'")
                    warnings = True
                    address = []

        return {'row_id': row_id, 'address': ','.join(address), 'original_columns': row, 'warnings': warnings}

    async def worker(self, csv_output, queue, progress):
        while True:
            item = await queue.get()

            try:
                await self.geocode_one_address(csv_output, item['row_id'], item['address'], item['original_columns'])

                if progress:
                    progress.update(1)
            except Exception as exc:
                traceback.print_exception(exc, file=sys.stderr)
            finally:
                queue.task_done()

    async def geocode_one_address(self, csv_output, row_id, address, original_columns):
        def on_backoff(details):
            if not self.options.quiet:
                sys.stderr.write("Backing off {wait:0.1f} seconds afters {tries} tries "
                                 "calling function {target} with args {args} and kwargs "
                                 "{kwargs}\n".format(**details))

        @backoff.on_exception(backoff.expo,
                              asyncio.TimeoutError,
                              max_time=self.options.timeout,
                              max_tries=self.options.retries,
                              on_backoff=on_backoff)
        async def _geocode_one_address():
            async with OpenCageGeocode(
                self.options.api_key,
                domain=self.options.api_domain,
                sslcontext=self.sslcontext,
                user_agent_comment=self.user_agent_comment
            ) as geocoder:
                geocoding_results = None
                response = None
                params = {'no_annotations': 1, 'raw_response': True, **self.options.optional_api_params}

                try:
                    if self.options.command == 'reverse':
                        if ',' in address:
                            lon, lat = address.split(',')
                            response = await geocoder.reverse_geocode_async(lon, lat, **params)
                            geocoding_results = floatify_latlng(response['results'])
                    else:
                        response = await geocoder.geocode_async(address, **params)
                        geocoding_results = floatify_latlng(response['results'])
                except OpenCageGeocodeError as exc:
                    self.log(str(exc))
                except Exception as exc:
                    traceback.print_exception(exc, file=sys.stderr)

                try:
                    if geocoding_results is not None and len(geocoding_results):
                        geocoding_result = geocoding_results[0]
                    else:
                        geocoding_result = None

                    if self.options.verbose:
                        self.log({
                            'row_id': row_id,
                            'thread_id': threading.get_native_id(),
                            'request': geocoder.url + '?' + urlencode(geocoder._parse_request(address, params)),
                            'response': response
                        })

                    await self.write_one_geocoding_result(
                        csv_output,
                        row_id,
                        geocoding_result,
                        response,
                        original_columns
                    )
                except Exception as exc:
                    traceback.print_exception(exc, file=sys.stderr)

        await _geocode_one_address()

    async def write_one_geocoding_result(
            self,
            csv_output,
            row_id,
            geocoding_result,
            raw_response,
            original_columns):
        row = original_columns

        for column in self.options.add_columns:
            if column == 'status':
                row.append(self.deep_get_result_value(raw_response, ['status', 'message']))
            elif geocoding_result is None:
                row.append('')
            elif column in geocoding_result:
                row.append(self.deep_get_result_value(geocoding_result, [column], ''))
            elif column in geocoding_result['components']:
                row.append(self.deep_get_result_value(geocoding_result, ['components', column], ''))
            elif column in geocoding_result['geometry']:
                row.append(self.deep_get_result_value(geocoding_result, ['geometry', column], ''))
            elif column == 'FIPS':
                row.append(
                    self.deep_get_result_value(
                        geocoding_result, [
                            'annotations', 'FIPS', 'county'], ''))
            elif column == 'json':
                row.append(json.dumps(geocoding_result, separators=(',', ':')))  # Compact JSON
            else:
                row.append('')

        # Enforce that row are written ordered. That means we might wait for other threads
        # to finish a task and make the overall process slower. Alternative would be to
        # use a second queue, or keep some results in memory.
        if not self.options.unordered:
            while row_id > self.write_counter:
                if self.options.verbose:
                    self.log(f"Want to write row {row_id}, but write_counter is at {self.write_counter}")
                await asyncio.sleep(random.uniform(0.01, 0.1))

            if self.options.verbose:
                self.log(f"Writing row {row_id}")
        csv_output.writerow(row)
        self.write_counter = self.write_counter + 1

    def log(self, message):
        if not self.options.quiet:
            sys.stderr.write(f"{message}\n")

    def deep_get_result_value(self, data, keys, default=None):
        for key in keys:
            if isinstance(data, dict):
                data = data.get(key, default)
            else:
                return default
        return data



import pytest

import responses

from opencage.geocoder import OpenCageGeocode
from opencage.geocoder import UnknownError

geocoder = OpenCageGeocode('abcde')


@responses.activate
def test_http_500_status():
    responses.add(
        responses.GET,
        geocoder.url,
        body='{}',
        status=500,
    )

    with pytest.raises(UnknownError) as excinfo:
        geocoder.geocode('whatever')

    assert str(excinfo.value) == '500 status code from API'


@responses.activate
def test_non_json():
    "These kinds of errors come from webserver and may not be JSON"
    responses.add(
        responses.GET,
        geocoder.url,
        body='<html><body><h1>503 Service Unavailable</h1></body></html>',
        headers={
            'Content-Type': 'text/html',
        },
        status=503
    )

    with pytest.raises(UnknownError) as excinfo:
        geocoder.geocode('whatever')

    assert str(excinfo.value) == 'Non-JSON result from server'


@responses.activate
def test_no_results_key():
    responses.add(
        responses.GET,
        geocoder.url,
        body='{"spam": "eggs"}',
        status=200,  # Need to specify status code with responses
    )

    with pytest.raises(UnknownError) as excinfo:
        geocoder.geocode('whatever')

    assert str(excinfo.value) == "JSON from API doesn't have a 'results' key"



""" Geocoder module. """

from decimal import Decimal
import collections

import os
import sys
import requests
import backoff
from .version import __version__

try:
    import aiohttp
    AIOHTTP_AVAILABLE = True
except ImportError:
    AIOHTTP_AVAILABLE = False

DEFAULT_DOMAIN = 'api.opencagedata.com'


def backoff_max_time():
    return int(os.environ.get('BACKOFF_MAX_TIME', '120'))


class OpenCageGeocodeError(Exception):

    """Base class for all errors/exceptions that can happen when geocoding."""


class InvalidInputError(OpenCageGeocodeError):

    """
    There was a problem with the input you provided.

    :var message: Error message describing the bad input.
    :var bad_value: The value that caused the problem.
    """

    def __init__(self, message, bad_value=None):
        super().__init__()
        self.message = message
        self.bad_value = bad_value

    def __unicode__(self):
        return self.message

    __str__ = __unicode__


class UnknownError(OpenCageGeocodeError):

    """There was a problem with the OpenCage server."""


class RateLimitExceededError(OpenCageGeocodeError):

    """
    Exception raised when account has exceeded it's limit.
    """

    def __unicode__(self):
        """Convert exception to a string."""
        return ("You have used the requests available on your plan. "
                "Please purchase more if you wish to continue: https://opencagedata.com/pricing")

    __str__ = __unicode__


class NotAuthorizedError(OpenCageGeocodeError):

    """
    Exception raised when an unautorized API key is used.
    """

    def __unicode__(self):
        """Convert exception to a string."""
        return "Your API key is not authorized. You may have entered it incorrectly."

    __str__ = __unicode__


class ForbiddenError(OpenCageGeocodeError):

    """
    Exception raised when a blocked or suspended API key is used.
    """

    def __unicode__(self):
        """Convert exception to a string."""
        return "Your API key has been blocked or suspended."

    __str__ = __unicode__


class AioHttpError(OpenCageGeocodeError):

    """
    Exceptions related to async HTTP calls with aiohttp
    """


class SSLError(OpenCageGeocodeError):

    """
    Exception raised when SSL connection to OpenCage server fails.
    """

    def __unicode__(self):
        """Convert exception to a string."""
        return ("SSL Certificate error connecting to OpenCage API. This is usually due to "
                "outdated CA root certificates of the operating system. "
                )

    __str__ = __unicode__


class OpenCageGeocode:

    """
    Geocoder object.

    Initialize it with your API key:

        >>> geocoder = OpenCageGeocode('your-key-here')

    Query:

        >>> geocoder.geocode("London")

    Reverse geocode a latitude & longitude into a place:

        >>> geocoder.reverse_geocode(51.5104, -0.1021)

    """

    session = None

    def __init__(
            self,
            key=None,
            protocol='https',
            domain=DEFAULT_DOMAIN,
            sslcontext=None,
            user_agent_comment=None):
        """Constructor."""
        self.key = key if key is not None else os.environ.get('OPENCAGE_API_KEY')

        if self.key is None:
            raise ValueError(
                "API key not provided. "
                "Either pass a 'key' parameter or set the OPENCAGE_API_KEY environment variable."
            )

        if protocol and protocol not in ('http', 'https'):
            protocol = 'https'
        self.url = protocol + '://' + domain + '/geocode/v1/json'

        # https://docs.aiohttp.org/en/stable/client_advanced.html#ssl-control-for-tcp-sockets
        self.sslcontext = sslcontext

        self.user_agent_comment = user_agent_comment

    def __enter__(self):
        self.session = requests.Session()
        return self

    def __exit__(self, *args):
        self.session.close()
        self.session = None
        return False

    async def __aenter__(self):
        if not AIOHTTP_AVAILABLE:
            raise AioHttpError("You must install `aiohttp` to use async methods")

        self.session = aiohttp.ClientSession()
        return self

    async def __aexit__(self, *args):
        await self.session.close()
        self.session = None
        return False

    def geocode(self, query, **kwargs):
        """
        Given a string to search for, return the list (array) of results from OpenCage's Geocoder.

        :param string query: String to search for

        :returns: Dict results
        :raises InvalidInputError: if the query string is not a unicode string
        :raises RateLimitExceededError: if you have exceeded the number of queries you can make.
        :                                  Exception says when you can try again
        :raises UnknownError: if something goes wrong with the OpenCage API

        """

        if self.session and isinstance(self.session, aiohttp.client.ClientSession):
            raise AioHttpError("Cannot use `geocode` in an async context, use `geocode_async`.")

        raw_response = kwargs.pop('raw_response', False)
        request = self._parse_request(query, kwargs)
        response = self._opencage_request(request)

        if raw_response:
            return response

        return floatify_latlng(response['results'])

    async def geocode_async(self, query, **kwargs):
        """
        Aync version of `geocode`.

        Given a string to search for, return the list (array) of results from OpenCage's Geocoder.

        :param string query: String to search for

        :returns: Dict results
        :raises InvalidInputError: if the query string is not a unicode string
        :raises RateLimitExceededError: if exceeded number of queries you can make. You can try again
        :raises UnknownError: if something goes wrong with the OpenCage API

        """

        if not AIOHTTP_AVAILABLE:
            raise AioHttpError("You must install `aiohttp` to use async methods.")

        if not self.session:
            raise AioHttpError("Async methods must be used inside an async context.")

        if not isinstance(self.session, aiohttp.client.ClientSession):
            raise AioHttpError("You must use `geocode_async` in an async context.")

        raw_response = kwargs.pop('raw_response', False)
        request = self._parse_request(query, kwargs)
        response = await self._opencage_async_request(request)

        if raw_response:
            return response

        return floatify_latlng(response['results'])

    def reverse_geocode(self, lat, lng, **kwargs):
        """
        Given a latitude & longitude, return an address for that point from OpenCage's Geocoder.

        :param lat: Latitude
        :param lng: Longitude
        :return: Results from OpenCageData
        :rtype: dict
        :raises RateLimitExceededError: if you have exceeded the number of queries you can make.
        :                                  Exception says when you can try again
        :raises UnknownError: if something goes wrong with the OpenCage API
        :raises InvalidInputError: if the latitude or longitude is out of bounds
        """

        self._validate_lat_lng(lat, lng)

        return self.geocode(_query_for_reverse_geocoding(lat, lng), **kwargs)

    async def reverse_geocode_async(self, lat, lng, **kwargs):
        """
        Aync version of `reverse_geocode`.

        Given a latitude & longitude, return an address for that point from OpenCage's Geocoder.

        :param lat: Latitude
        :param lng: Longitude
        :return: Results from OpenCageData
        :rtype: dict
        :raises RateLimitExceededError: if exceeded number of queries you can make. You can try again
        :raises UnknownError: if something goes wrong with the OpenCage API
        :raises InvalidInputError: if the latitude or longitude is out of bounds
        """

        self._validate_lat_lng(lat, lng)

        return await self.geocode_async(_query_for_reverse_geocoding(lat, lng), **kwargs)

    @backoff.on_exception(
        backoff.expo,
        (UnknownError, requests.exceptions.RequestException),
        max_tries=5, max_time=backoff_max_time)
    def _opencage_request(self, params):

        if self.session:
            response = self.session.get(self.url, params=params, headers=self._opencage_headers('aiohttp'))
        else:
            response = requests.get(self.url, params=params, headers=self._opencage_headers('requests'))

        try:
            response_json = response.json()
        except ValueError as excinfo:
            raise UnknownError("Non-JSON result from server") from excinfo

        if response.status_code == 401:
            raise NotAuthorizedError()

        if response.status_code == 403:
            raise ForbiddenError()

        if response.status_code in (402, 429):
            raise RateLimitExceededError()

        if response.status_code == 500:
            raise UnknownError("500 status code from API")

        if 'results' not in response_json:
            raise UnknownError("JSON from API doesn't have a 'results' key")

        return response_json

    def _opencage_headers(self, client):
        client_version = requests.__version__
        if client == 'aiohttp':
            client_version = aiohttp.__version__

        py_version = '.'.join(str(x) for x in sys.version_info[0:3])

        comment = ''
        if self.user_agent_comment:
            comment = f" ({self.user_agent_comment})"

        return {
            'User-Agent': f"opencage-python/{__version__} Python/{py_version} {client}/{client_version}{comment}"
        }

    async def _opencage_async_request(self, params):
        try:
            async with self.session.get(self.url, params=params, ssl=self.sslcontext) as response:
                try:
                    response_json = await response.json()
                except ValueError as excinfo:
                    raise UnknownError("Non-JSON result from server") from excinfo

                if response.status == 401:
                    raise NotAuthorizedError()

                if response.status == 403:
                    raise ForbiddenError()

                if response.status in (402, 429):
                    raise RateLimitExceededError()

                if response.status == 500:
                    raise UnknownError("500 status code from API")

                if 'results' not in response_json:
                    raise UnknownError("JSON from API doesn't have a 'results' key")

                return response_json
        except aiohttp.ClientSSLError as exp:
            raise SSLError() from exp
        except aiohttp.client_exceptions.ClientConnectorCertificateError as exp:
            raise SSLError() from exp

    def _parse_request(self, query, params):
        if not isinstance(query, str):
            error_message = "Input must be a unicode string, not " + repr(query)[:100]
            raise InvalidInputError(error_message, bad_value=query)

        data = {'q': query, 'key': self.key}
        data.update(params)  # Add user parameters
        return data

    def _validate_lat_lng(self, lat, lng):
        """
        Validate latitude and longitude values.

        Raises InvalidInputError if the values are out of bounds.
        """
        try:
            lat_float = float(lat)
            if not -90 <= lat_float <= 90:
                raise InvalidInputError(f"Latitude must be a number between -90 and 90, not {lat}", bad_value=lat)
        except ValueError:
            raise InvalidInputError(f"Latitude must be a number between -90 and 90, not {lat}", bad_value=lat)

        try:
            lng_float = float(lng)
            if not -180 <= lng_float <= 180:
                raise InvalidInputError(f"Longitude must be a number between -180 and 180, not {lng}", bad_value=lng)
        except ValueError:
            raise InvalidInputError(f"Longitude must be a number between -180 and 180, not {lng}", bad_value=lng)


def _query_for_reverse_geocoding(lat, lng):
    """
    Given a lat & lng, what's the string search query.

    If the API changes, change this function. Only for internal use.
    """
    # have to do some stupid f/Decimal/str stuff to (a) ensure we get as much
    # decimal places as the user already specified and (b) to ensure we don't
    # get e-5 stuff
    return f"{Decimal(str(lat)):f},{Decimal(str(lng)):f}"


def float_if_float(float_string):
    """
    Given a float string, returns the float value.
    On value error returns the original string.
    """
    try:
        float_val = float(float_string)
        return float_val
    except ValueError:
        return float_string


def floatify_latlng(input_value):
    """
    Work around a JSON dict with string, not float, lat/lngs.

    Given anything (list/dict/etc) it will return that thing again, *but* any
    dict (at any level) that has only 2 elements lat & lng, will be replaced
    with the lat & lng turned into floats.

    If the API returns the lat/lng as strings, and not numbers, then this
    function will 'clean them up' to be floats.
    """
    if isinstance(input_value, collections.abc.Mapping):
        if len(input_value) == 2 and sorted(input_value.keys()) == ['lat', 'lng']:
            # This dict has only 2 keys 'lat' & 'lon'
            return {'lat': float_if_float(input_value["lat"]), 'lng': float_if_float(input_value["lng"])}

        return dict((key, floatify_latlng(value)) for key, value in input_value.items())

    if isinstance(input_value, collections.abc.MutableSequence):
        return [floatify_latlng(x) for x in input_value]

    return input_value



# encoding: utf-8

from pathlib import Path

import os
import re
import responses

from opencage.geocoder import OpenCageGeocode

# reduce maximum backoff retry time from 120s to 1s
os.environ['BACKOFF_MAX_TIME'] = '1'

geocoder = OpenCageGeocode('abcde', user_agent_comment='OpenCage Test')

user_agent_format = re.compile(
    r'^opencage-python/[\d\.]+ Python/[\d\.]+ (requests|aiohttp)/[\d\.]+ \(OpenCage Test\)$')


@responses.activate
def test_sync():
    responses.add(
        responses.GET,
        geocoder.url,
        body=Path('test/fixtures/uk_postcode.json').read_text(encoding="utf-8"),
        status=200
    )

    geocoder.geocode("EC1M 5RF")

    # Check the User-Agent header in the most recent request
    request = responses.calls[-1].request
    user_agent = request.headers['User-Agent']

    assert user_agent_format.match(user_agent) is not None



# encoding: utf-8

from pathlib import Path

import os
import responses

from opencage.geocoder import OpenCageGeocode

# reduce maximum backoff retry time from 120s to 1s
os.environ['BACKOFF_MAX_TIME'] = '1'


geocoder = OpenCageGeocode('abcde')


def _any_result_around(results, lat=None, lon=None):
    for result in results:
        if (abs(result['geometry']['lat'] - lat) < 0.05
                and abs(result['geometry']['lng'] - lon) < 0.05):
            return True
    return False


@responses.activate
def test_gb_postcode():
    responses.add(
        responses.GET,
        geocoder.url,
        body=Path('test/fixtures/uk_postcode.json').read_text(encoding="utf-8"),
        status=200
    )

    results = geocoder.geocode("EC1M 5RF")
    assert _any_result_around(results, lat=51.5201666, lon=-0.0985142)


@responses.activate
def test_australia():
    responses.add(
        responses.GET,
        geocoder.url,
        body=Path('test/fixtures/mudgee_australia.json').read_text(encoding="utf-8"),
        status=200
    )

    results = geocoder.geocode("Mudgee, Australia")
    assert _any_result_around(results, lat=-32.5980702, lon=149.5886383)


@responses.activate
def test_munster():
    responses.add(
        responses.GET,
        geocoder.url,
        body=Path('test/fixtures/muenster.json').read_text(encoding="utf-8"),
        status=200
    )

    results = geocoder.geocode("Münster")
    assert _any_result_around(results, lat=51.9625101, lon=7.6251879)


@responses.activate
def test_donostia():
    responses.add(
        responses.GET,
        geocoder.url,
        body=Path('test/fixtures/donostia.json').read_text(encoding="utf-8"),
        status=200
    )

    results = geocoder.geocode("Donostia")
    assert _any_result_around(results, lat=43.300836, lon=-1.9809529)

    # test that the results are in unicode
    assert results[0]['formatted'] == 'San Sebastián, Autonomous Community of the Basque Country, Spain'



import pytest

import responses

from opencage.geocoder import OpenCageGeocode
from opencage.geocoder import InvalidInputError

geocoder = OpenCageGeocode('abcde')


@responses.activate
def test_must_be_unicode_string():
    responses.add(
        responses.GET,
        geocoder.url,
        body='{"results":{}}',
        status=200
    )

    # Should not give errors
    geocoder.geocode('xxx')   # ascii convertable
    geocoder.geocode('xxá')   # unicode

    # But if it isn't a unicode string, it should give error
    utf8_string = "xxá".encode("utf-8")
    latin1_string = "xxá".encode("latin1")

    with pytest.raises(InvalidInputError) as excinfo:
        geocoder.geocode(utf8_string)
    assert str(excinfo.value) == f"Input must be a unicode string, not {utf8_string!r}"
    assert excinfo.value.bad_value == utf8_string

    with pytest.raises(InvalidInputError) as excinfo:
        geocoder.geocode(latin1_string)
    assert str(excinfo.value) == f"Input must be a unicode string, not {latin1_string!r}"
    assert excinfo.value.bad_value == latin1_string


@responses.activate
def test_reject_out_of_bounds_coordinates():
    """Test that reverse geocoding rejects out-of-bounds latitude and longitude values."""
    responses.add(
        responses.GET,
        geocoder.url,
        body='{"results":{}}',
        status=200
    )

    # Valid coordinates should work
    geocoder.reverse_geocode(45.0, 90.0)
    geocoder.reverse_geocode(-45.0, -90.0)

    # Invalid latitude values (outside -90 to 90)
    with pytest.raises(InvalidInputError) as excinfo:
        geocoder.reverse_geocode(91.0, 45.0)
    assert "Latitude must be a number between -90 and 90" in str(excinfo.value)

    with pytest.raises(InvalidInputError) as excinfo:
        geocoder.reverse_geocode(-91.0, 45.0)
    assert "Latitude must be a number between -90 and 90" in str(excinfo.value)

    # Invalid longitude values (outside -180 to 180)
    with pytest.raises(InvalidInputError) as excinfo:
        geocoder.reverse_geocode(45.0, 181.0)
    assert "Longitude must be a number between -180 and 180" in str(excinfo.value)

    with pytest.raises(InvalidInputError) as excinfo:
        geocoder.reverse_geocode(45.0, -181.0)
    assert "Longitude must be a number between -180 and 180" in str(excinfo.value)



from pathlib import Path

import pytest
import responses

from opencage.geocoder import OpenCageGeocode
from opencage.geocoder import NotAuthorizedError

geocoder = OpenCageGeocode('unauthorized-key')


@responses.activate
def test_api_key_not_authorized():
    responses.add(
        responses.GET,
        geocoder.url,
        body=Path('test/fixtures/401_not_authorized.json').read_text(encoding="utf-8"),
        status=401,
    )

    with pytest.raises(NotAuthorizedError) as excinfo:
        geocoder.geocode("whatever")
    assert str(excinfo.value) == 'Your API key is not authorized. You may have entered it incorrectly.'



from pathlib import Path

import pytest
import responses

from opencage.geocoder import OpenCageGeocode
from opencage.geocoder import ForbiddenError


geocoder = OpenCageGeocode('2e10e5e828262eb243ec0b54681d699a')  # will always return 403


@responses.activate
def test_api_key_blocked():
    responses.add(
        responses.GET,
        geocoder.url,
        body=Path('test/fixtures/403_apikey_disabled.json').read_text(encoding="utf-8"),
        status=403,
    )

    with pytest.raises(ForbiddenError) as excinfo:
        geocoder.geocode("whatever")
    assert str(excinfo.value) == 'Your API key has been blocked or suspended.'



from opencage.geocoder import floatify_latlng


def test_string():
    assert floatify_latlng("123") == "123"


def test_empty_dict():
    assert floatify_latlng({}) == {}


def test_empty_list():
    assert floatify_latlng([]) == []


def test_dict_with_floats():
    assert floatify_latlng({'geom': {'lat': 12.01, 'lng': -0.9}}) == {'geom': {'lat': 12.01, 'lng': -0.9}}


def dict_with_stringified_floats():
    assert floatify_latlng({'geom': {'lat': "12.01", 'lng': "-0.9"}}) == {'geom': {'lat': 12.01, 'lng': -0.9}}


def dict_with_list():
    assert floatify_latlng(
        {'results': [{'geom': {'lat': "12.01", 'lng': "-0.9"}}, {'geometry': {'lat': '0.1', 'lng': '10'}}]}
    ) == {'results': [{'geom': {'lat': 12.01, 'lng': -0.9}}, {'geometry': {'lat': 0.1, 'lng': 10}}]}


def list_with_things():
    assert floatify_latlng([{'foo': 'bar'}]) == [{'foo': 'bar'}]



# encoding: utf-8

import ssl
import pytest
from opencage.geocoder import OpenCageGeocode, SSLError

# NOTE: Testing keys https://opencagedata.com/api#testingkeys

# Connect to a host that has an invalid certificate


@pytest.mark.asyncio
async def test_sslerror():
    bad_domain = 'wrong.host.badssl.com'
    async with OpenCageGeocode('6d0e711d72d74daeb2b0bfd2a5cdfdba', domain=bad_domain) as geocoder:
        with pytest.raises(SSLError) as excinfo:
            await geocoder.geocode_async("something")
        assert str(excinfo.value).startswith('SSL Certificate error')

# Connect to OpenCage API domain but use certificate of another domain
# This tests that sslcontext can be set.


@pytest.mark.asyncio
async def test_sslerror_wrong_certificate():
    sslcontext = ssl.create_default_context(cafile='test/fixtures/badssl-com-chain.pem')

    async with OpenCageGeocode('6d0e711d72d74daeb2b0bfd2a5cdfdba', sslcontext=sslcontext) as geocoder:
        with pytest.raises(SSLError) as excinfo:
            await geocoder.geocode_async("something")
        assert str(excinfo.value).startswith('SSL Certificate error')



# encoding: utf-8

import pytest

from opencage.geocoder import ForbiddenError, OpenCageGeocode, AioHttpError

# NOTE: Testing keys https://opencagedata.com/api#testingkeys


@pytest.mark.asyncio
async def test_success():
    async with OpenCageGeocode('6d0e711d72d74daeb2b0bfd2a5cdfdba') as geocoder:
        results = await geocoder.geocode_async("EC1M 5RF")
        assert any(
            abs(result['geometry']['lat'] - 51.952659 < 0.05
                and abs(result['geometry']['lng'] - 7.632473) < 0.05)
            for result in results
        )


@pytest.mark.asyncio
async def test_failure():
    async with OpenCageGeocode('6c79ee8e1ca44ad58ad1fc493ba9542f') as geocoder:
        with pytest.raises(ForbiddenError) as excinfo:
            await geocoder.geocode_async("Atlantis")

        assert str(excinfo.value) == 'Your API key has been blocked or suspended.'


@pytest.mark.asyncio
async def test_without_async_session():
    geocoder = OpenCageGeocode('4372eff77b8343cebfc843eb4da4ddc4')

    with pytest.raises(AioHttpError) as excinfo:
        await geocoder.geocode_async("Atlantis")

    assert str(excinfo.value) == 'Async methods must be used inside an async context.'


@pytest.mark.asyncio
async def test_using_non_async_method():
    async with OpenCageGeocode('6d0e711d72d74daeb2b0bfd2a5cdfdba') as geocoder:
        with pytest.raises(AioHttpError) as excinfo:
            await geocoder.geocode("Atlantis")

    assert str(excinfo.value) == 'Cannot use `geocode` in an async context, use `geocode_async`.'



# encoding: utf-8

from pathlib import Path

import pytest
import responses

from opencage.geocoder import OpenCageGeocode
from opencage.geocoder import NotAuthorizedError


def _any_result_around(results, lat=None, lon=None):
    for result in results:
        if (abs(result['geometry']['lat'] - lat) < 0.05
                and abs(result['geometry']['lng'] - lon) < 0.05):
            return True
    return False


@responses.activate
def test_success():
    with OpenCageGeocode('abcde') as geocoder:
        responses.add(
            responses.GET,
            geocoder.url,
            body=Path('test/fixtures/uk_postcode.json').read_text(encoding="utf-8"),
            status=200
        )

        results = geocoder.geocode("EC1M 5RF")
        assert _any_result_around(results, lat=51.5201666, lon=-0.0985142)


@responses.activate
def test_failure():
    with OpenCageGeocode('unauthorized-key') as geocoder:
        responses.add(
            responses.GET,
            geocoder.url,
            body=Path('test/fixtures/401_not_authorized.json').read_text(encoding="utf-8"),
            status=401,
        )

        with pytest.raises(NotAuthorizedError) as excinfo:
            geocoder.geocode("whatever")
        assert str(excinfo.value) == 'Your API key is not authorized. You may have entered it incorrectly.'



from pathlib import Path

import pytest
import responses

from opencage.geocoder import OpenCageGeocode
from opencage.geocoder import RateLimitExceededError

geocoder = OpenCageGeocode('abcde')


@responses.activate
def test_no_rate_limit():
    responses.add(
        responses.GET,
        geocoder.url,
        body=Path('test/fixtures/no_ratelimit.json').read_text(encoding="utf-8"),
        status=200
    )
    # shouldn't raise an exception
    geocoder.geocode("whatever")


@responses.activate
def test_rate_limit_exceeded():
    # 4372eff77b8343cebfc843eb4da4ddc4 will always return 402
    responses.add(
        responses.GET,
        geocoder.url,
        body=Path('test/fixtures/402_rate_limit_exceeded.json').read_text(encoding="utf-8"),
        status=402,
        headers={
            'X-RateLimit-Limit': '2500',
            'X-RateLimit-Remaining': '0',
            'X-RateLimit-Reset': '1402185600'
        }
    )

    with pytest.raises(RateLimitExceededError) as excinfo:
        geocoder.geocode("whatever")
    assert 'You have used the requests available on your plan.' in str(excinfo.value)



#!/usr/bin/env python

import os
import sys
from setuptools import setup, find_packages

# if you are not using vagrant, just delete os.link directly,
# The hard link only saves a little disk space, so you should not care
if os.environ.get('USER', '') == 'vagrant':
    del os.link


ROOT_DIR = os.path.dirname(__file__)
SOURCE_DIR = os.path.join(ROOT_DIR)

if sys.version_info < (3, 8):
    raise RuntimeError(
        "opencage requires Python 3.8 or newer"
        "Use older opencage 1.x for Python 2.7 or 3.7"
    )

# try for testing
try:
    with open(os.path.join(SOURCE_DIR, 'README.md'), encoding="utf-8") as f:
        LONG_DESCRIPTION = f.read()
except FileNotFoundError:
    LONG_DESCRIPTION = ""

setup(
    name="opencage",
    version="3.2.0",
    description="Wrapper module for the OpenCage Geocoder API",
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author="OpenCage GmbH",
    author_email="info@opencagedata.com",
    url="https://github.com/OpenCageData/python-opencage-geocoder/",
    download_url="https://github.com/OpenCageData/python-opencage-geocoder/tarball/3.2.0",
    license="BSD",
    entry_points={
        'console_scripts': [
            'opencage=opencage.command_line:main'
        ]
    },
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    keywords=['geocoding', 'geocoder'],
    classifiers=[
        'Environment :: Web Environment',
        "Development Status :: 5 - Production/Stable",
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        "Programming Language :: Python :: 3 :: Only",
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Utilities'
    ],
    install_requires=[
        'Requests>=2.31.0',
        'backoff>=2.2.1',
        'tqdm>=4.66.4',
        'certifi>=2024.07.04',
        'aiohttp>=3.10.5'
    ],
    test_suite='pytest',
    tests_require=[
        'responses>=0.25.7',
        'flake8>=7.0.0',
        'pytest>=7.4.0'
    ],
)



from opencage.batch import OpenCageBatchGeocoder

batch = OpenCageBatchGeocoder({})


def test_deep_get_result_value():
    result = {
        'annotations': {
            'FIPS': {
                'state': 'CA'
            }
        },
        'components': {
            'street': 'Main Road'
        }
    }

    assert batch.deep_get_result_value(result, ['hello', 'world']) is None

    assert batch.deep_get_result_value(result, ['components', 'street']) == 'Main Road'
    assert batch.deep_get_result_value(result, ['components', 'city']) is None
    assert batch.deep_get_result_value(result, ['components', 'city'], '') == ''

    assert batch.deep_get_result_value([], ['hello', 'world']) is None
    assert batch.deep_get_result_value(None, ['hello', 'world']) is None



# encoding: utf-8

from opencage.geocoder import OpenCageGeocode

import os


def test_protocol_http():
    """Test that HTTP protocol can be set correctly"""
    geocoder = OpenCageGeocode('abcde', protocol='http')
    assert geocoder.url == 'http://api.opencagedata.com/geocode/v1/json'


def test_api_key_env_var():
    """Test that API key can be set by an environment variable"""

    os.environ['OPENCAGE_API_KEY'] = 'from-env-var'
    geocoder = OpenCageGeocode()
    assert geocoder.key == 'from-env-var'


def test_custom_domain():
    """Test that custom domain can be set"""
    geocoder = OpenCageGeocode('abcde', domain='example.com')
    assert geocoder.url == 'https://example.com/geocode/v1/json'



import argparse
import sys
import os
import io
import re
import csv

from opencage.batch import OpenCageBatchGeocoder
from opencage.version import __version__


def main(args=sys.argv[1:]):
    options = parse_args(args)

    assert sys.version_info >= (3, 8), "Script requires Python 3.8 or newer"

    geocoder = OpenCageBatchGeocoder(options)

    with options.input as input_filename:
        with (io.StringIO() if options.dry_run else open(options.output, 'x', encoding='utf-8')) as output_io:
            reader = csv.reader(input_filename, strict=True, skipinitialspace=True)
            writer = csv.writer(output_io)

            geocoder(csv_input=reader, csv_output=writer)


def parse_args(args):
    if len(args) == 0:
        print("To display help use 'opencage -h', 'opencage forward -h' or 'opencage reverse -h'", file=sys.stderr)
        sys.exit(1)

    parser = argparse.ArgumentParser(description=f'Opencage CLI {__version__}')
    parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}')

    subparsers = parser.add_subparsers(dest='command')
    subparsers.required = True

    subparser_forward = subparsers.add_parser(
        'forward', help="Forward geocode a file (input is address, add coordinates)")
    subparser_reverse = subparsers.add_parser(
        'reverse', help="Reverse geocode a file (input is coordinates, add full address)")

    for subparser in [subparser_forward, subparser_reverse]:
        subparser.add_argument("--api-key", required=True, type=api_key_type, help="Your OpenCage API key")
        subparser.add_argument(
            "--input",
            required=True,
            type=argparse.FileType(
                'r',
                encoding='utf-8'),
            help="Input file name",
            metavar='FILENAME')
        subparser.add_argument(
            "--output",
            required=True,
            type=str,
            help="Output file name",
            metavar='FILENAME')

        add_optional_arguments(subparser)

    options = parser.parse_args(args)

    if os.path.exists(options.output) and not options.dry_run:
        if options.overwrite:
            os.remove(options.output)
        else:
            print(
                f"Error: The output file '{options.output}' already exists. You can add --overwrite to your command.",
                file=sys.stderr)
            sys.exit(1)

    if 0 in options.input_columns:
        print("Error: A column 0 in --input-columns does not exist. The lowest possible number is 1.", file=sys.stderr)
        sys.exit(1)

    return options


def add_optional_arguments(parser):
    parser.add_argument(
        "--headers",
        action="store_true",
        help="If the first row should be treated as a header row")
    default_input_cols = '1,2' if re.match(r'.*reverse', parser.prog) else '1'
    parser.add_argument(
        "--input-columns",
        type=comma_separated_type(int),
        default=default_input_cols,
        help=f"Comma-separated list of integers (default '{default_input_cols}')",
        metavar='')
    default_add_cols = (
        'lat,lng,_type,_category,country_code,country,state,county,_normalized_city,'
        'postcode,road,house_number,confidence,formatted'
    )
    parser.add_argument(
        "--add-columns",
        type=comma_separated_type(str),
        default=default_add_cols,
        help=f"Comma-separated list of output columns (default '{default_add_cols}')",
        metavar='')
    parser.add_argument("--workers", type=ranged_type(int, 1, 20), default=1,
                        help="Number of parallel geocoding requests (default 1)", metavar='')
    parser.add_argument("--timeout", type=ranged_type(int, 1, 60), default=10,
                        help="Timeout in seconds (default 10)", metavar='')
    parser.add_argument("--retries", type=ranged_type(int, 1, 60), default=10,
                        help="Number of retries (default 5)", metavar='')
    parser.add_argument("--api-domain", type=str, default="api.opencagedata.com",
                        help="API domain (default api.opencagedata.com)", metavar='')
    parser.add_argument("--optional-api-params", type=comma_separated_dict_type, default="",
                        help="Extra parameters for each request (e.g. language=fr,no_dedupe=1)", metavar='')
    parser.add_argument(
        "--limit",
        type=int,
        default=0,
        help="Stop after this number of lines in the input",
        metavar='')
    parser.add_argument(
        "--unordered",
        action="store_true",
        help="Allow the output lines to be in different order (can be faster)")
    parser.add_argument("--dry-run", action="store_true", help="Read the input file but no geocoding")
    parser.add_argument("--no-progress", action="store_true", help="Display no progress bar")
    parser.add_argument("--quiet", action="store_true", help="No progress bar and no messages")
    parser.add_argument("--overwrite", action="store_true", help="Delete the output file first if it exists")
    parser.add_argument("--verbose", action="store_true", help="Display debug information for each request")

    return parser


def api_key_type(apikey):
    pattern = re.compile(r"^(oc_gc_)?[0-9a-f]{32}$")

    if not pattern.match(apikey):
        raise argparse.ArgumentTypeError("invalid API key")

    return apikey


def ranged_type(value_type, min_value, max_value):
    def range_checker(arg: str):
        try:
            f = value_type(arg)
        except ValueError as exc:
            raise argparse.ArgumentTypeError(f'must be a valid {value_type}') from exc
        if f < min_value or f > max_value:
            raise argparse.ArgumentTypeError(f'must be within [{min_value}, {max_value}]')
        return f

    # Return function handle to checking function
    return range_checker


def comma_separated_type(value_type):
    def comma_separated(arg: str):
        if not arg:
            return []

        return [value_type(x) for x in arg.split(',')]

    return comma_separated


def comma_separated_dict_type(arg):
    if not arg:
        return {}

    try:
        return dict([x.split('=') for x in arg.split(',')])
    except ValueError as exc:
        raise argparse.ArgumentTypeError("must be a valid comma separated list of key=value pairs") from exc
